;;;; -*- Mode: LISP; Syntax: Common-Lisp; Package: STANDARD-LISP; Base: 10 -*-
;;;
;;; STANDARD-LISP, Module SLCP
;;;
;;; ***************************************************************************
;;;
;;;        (c) Copyright 1985,1986,1987 by Lucid Inc.
;;;        Unrestricted permission is granted to copy, modify
;;;        or redistribute this file.
;;;
;;; ***************************************************************************
;;;
;;; The functions in the Standard Lisp Report.
;;;
;;; Programmer: Eric Benson
;;;
;;; Edit-History:
;;;
;;; Created: Saturday the twentieth of July, 1985; 8:45:20 pm
;;; 27-Apr-86 eb: Changed MAPC and MAP to return NIL, instead of the list
;;;		  argument as in Common Lisp.
;;; 28-Jun-86 eb: Changed QUOTIENT to be correct for floating point.
;;; 28-Jun-86 eb: Reorganized inline compilation, added VaxLisp feature.
;;;  9-Feb-87 eb: Changed INTERN to just call LISP:INTERN, not try to
;;;		  import the symbol.
;;; 29-Apr-87 eb: Added definition for REMPROP.
;;;
;;; End-of-Edit-History

(in-package 'standard-lisp)

;;; Set up our compilation environment

;;; This makes it easier to refer to DEC's Vax Lisp without confusing it
;;; with NIL or Lucid Common Lisp on a Vax.
#+(and Vax Common)
(eval-when (eval compile)
  (push 'VaxLisp *features*))

#+KCL
(eval-when (eval compile)
  (when (macro-function 'si::define-inline-function)
    (push :NewKCL *features*)))

#+Lucid
(eval-when (eval compile)
  (if (fboundp 'lucid::proclaim-unspecial)
      (push :NewLucidSpecials *features*)
      (push :OldLucidSpecials *features*)))

#+Symbolics
(eval-when (eval compile)
  (when (< (or (si:get-release-version) (si:get-system-version)) 7)
    (push :OldSymbolics *features*)))
    

;;; The following definitions aid in defining Standard Lisp functions so
;;; that they can be compiled efficiently.

#+Excl
(defmacro defsubst (fn arg &body body)
  `(defmacro ,fn ,arg
	     (cond ((or ,@(do ((v arg (cdr v))
			       (r nil (cons `(consp ,(car v)) r)))
			      ((null v) (nreverse r))))
		    (copy-list
		     (list '(lambda ,arg ,(if (null (cdr body))
					      (car body)
					      `(progn . ,body)))
			   ,@arg)))
		   (t
		    (lisp:sublis (list ,@(do ((v arg (cdr v))
					      (r nil (cons `(cons
							     ',(car v)
							     ,(car v))
							   r)))
					     ((null v) (nreverse r))))
				 (macroexpand ',(if (null (cdr body))
						    (car body)
						 `(progn . ,body))))))))

;;; Functions which should be compiled inline with proper functional
;;; semantics (evaluation order and single evaluation of arguments) are
;;; defined with this.
(defmacro define-inline (name arglist &body body)
  #+(or Lucid LispM Excl)
  `(defsubst ,name ,arglist ,@body)
  #+NewKCL
  `(si:define-inline-function ,name ,arglist ,@body)
  #-(or Lucid LispM NewKCL Excl)
  `(progn (proclaim '(inline ,name))
	  (defun ,name ,arglist ,@body)))
  
;;; This defines a macro for use only in the compiler.  Not all systems
;;; have this capability.  In transporting this program to a new Lisp
;;; implementation, it is wise to take advantage of this functionality
;;; if it exists.  This is especially important for compiling calls to
;;; MAPC, MAPCAR, etc, which might otherwise require a heap-consed funarg.
(defmacro define-compiler-macro (name arglist &body body)
  #+Lucid
  `(eval-when (eval compile load)
     (lucid::def-compiler-macro ,name ,arglist ,@body))
  #+VaxLisp
  `(eval-when (eval compile load)
     (system::deftransform ,name ,arglist ,@body))
  #+NewKCL
  `(eval-when (eval compile load)
     (si:define-compiler-macro ,name ,arglist ,@body))
  #+Excl
  `(eval-when (eval compile load)
     (excl::defcmacro ,name ,arglist ,@body))
  #+Symbolics
  (let ((optimizer-name (lisp:intern (concatenate 'string
						  (symbol-name name)
						  "-COMPILER-MACRO"))))
    `(eval-when (eval compile load)
       (defun ,optimizer-name (.form.)
	 (scl:destructuring-bind ,arglist (rest .form.)
	   ,@body))
       (compiler:add-optimizer ,name ,optimizer-name)))
  #-(or Lucid VaxLisp Symbolics NewKCL Excl)
  (error "Compiler macros have not been set up in this version of Lisp")
  )

;;; Functions which can be compiled more efficiently with a macro
;;; definition which does not have proper functional semantics, but also
;;; need to be functions, are defined with this.  If the capability of
;;; defining compiler macros has not been defined for this system, then
;;; DEFINE-INLINE is used instead.
(defmacro define-compiler-macro-and-function (name arglist &body body)
  #+(or Lucid VaxLisp Symbolics NewKCL Excl)
  `(progn
     (define-compiler-macro ,name ,arglist ,@body)
     (defun ,name ,arglist
       (macrolet ((.local-macro. ,arglist
		    ,@body))
	 (.local-macro. ,@arglist))))
  #-(or Lucid VaxLisp Symbolics NewKCL Excl)
  `(define-inline ,name ,arglist
     (macrolet ((.local-macro. ,arglist
		  ,@body))
       (.local-macro. ,@arglist))))

;;; Functions with a simple renaming are defined with this.
(defmacro define-synonym (new-name old-name &optional (arglist '(x)))
  `(define-compiler-macro-and-function ,new-name ,arglist
     (list ',old-name ,@arglist)))

;;; Functions with two arguments reversed are defined with this
(defmacro define-reversed-arguments (new-name old-name)
  `(define-compiler-macro-and-function ,new-name (x y)
     (list ',old-name y x)))
  

;;; Functions which are identical in Common Lisp and
;;; Standard Lisp are listed here with their names preceded by a
;;; single semicolon.

;3.1
;atom

(define-synonym codep compiled-function-p)

;;;Common Lisp CONSTANTP is incompatible with Standard Lisp
(define-inline constantp (x)
  (not (or (consp x) (symbolp x))))

;eq

(define-synonym eqn eql (x y))

;;;Common Lisp EQUAL does not compare vector contents
(defun equal (a b)
  (cond ((eq a b) t)
	((or (symbolp a) (symbolp b)) nil)
	((consp a)
	 (and (consp b)
	      (equal (car a) (car b))
	      (equal (cdr a) (cdr b))))
	((consp b)
	 nil)
	((numberp a)
	 (and (numberp b)
	      (eql a b)))
	((numberp b)
	 nil)
	((stringp a)
	 (and (stringp b)
	      (string= a b)))
	((stringp b)
	 nil)
	((simple-vector-p a)
	 (and (simple-vector-p b)
	      (let ((a-length (lisp:length a)))
		(and (eql a-length (lisp:length b))
		     (dotimes (i a-length t)
		       (unless (equal (svref a i) (svref b i))
			 (return nil)))))))
	(t nil)))

(define-synonym fixp integerp)

;floatp

(define-synonym idp symbolp)

;;;Standard Lisp MINUSP is more permissive than Common Lisp.
(define-inline minusp (x)
  (and (numberp x) (lisp:minusp x)))

;null

;numberp

(define-inline onep (x)
  (and (numberp x) (= x 1)))

(define-synonym pairp consp)

;stringp

;;;VECTORP is only true of things made by MKVECT.
(define-inline vectorp (x)
  (simple-vector-p x))

;;;Standard Lisp ZEROP is more permissive than Common Lisp.
(define-inline zerop (x)
  (and (numberp x) (lisp:zerop x)))

;3.2

;car,cdr,c...r,cons,list,rplaca,rplacd


;3.3

(defun compress (x)
  (let ((1st (first x)))
    (cond ((eq 1st '\")
	   (setq x (rest x))
	   ;;Make a string
	   (let ((len 0)
		 (previous-quote nil))
	     (declare (fixnum len))
	     ;;First find out how long it's going to be.
	     (dolist (ch x)
	       (if (eq ch '\")
		   (if previous-quote
		       (progn (incf len)
			      (setq previous-quote nil))
		       (setq previous-quote t))
		   (incf len)))
	     ;;Then create the string and put the characters in it.
	     (let ((str (make-string len))
		   (i 0))
	       (declare (fixnum i))
	       (setq previous-quote nil)
	       (dolist (ch x str)
		 (if (eq ch '\")
		     (if previous-quote
			 (progn (setf (schar str i) (char (symbol-name ch) 0))
				(incf i)
				(setq previous-quote nil))
			 (setq previous-quote t))
		     (progn
		       (setf (schar str i) (char (symbol-name ch) 0))
		       (incf i)))))))
	  ((or (digit 1st)
	       (eq 1st '+)
	       (eq 1st '-)
	       (eq 1st '\.))
	   ;;Make a number by reading from a string
	   (let ((*readtable* *standard-lisp-readtable*))
	     (read-from-string
	       (lisp:map 'string
			 #'(lambda (ch) (char (symbol-name ch) 0))
			 x))))
	  (t
	   ;;Make an uninterned symbol
	   (let ((len 0)
		 (previous-quote nil))
	     (declare (fixnum len))
	     ;;Figure out how long the symbol name will be.
	     (dolist (ch x)
	       (if previous-quote
		   (progn (incf len)
			  (setq previous-quote nil))
		   (if (eq ch '!)
		       (setq previous-quote t)
		       (incf len))))
	     ;;Create the string and make a symbol with that name.
	     (let ((str (make-string len))
		   (i 0))
	       (declare (fixnum i))
	       (setq previous-quote nil)
	       (dolist (ch x (make-symbol str))
		 (if previous-quote
		     (progn (setf (schar str i) (char (symbol-name ch) 0))
			    (incf i)
			    (setq previous-quote nil))
		     (if (eq ch '!)
			 (setq previous-quote t)
			 (progn
			   (setf (schar str i) (char (symbol-name ch) 0))
			   (incf i)))))))))))

(defun explode (x)
  (explode1 x t))

;;;Not Standard Lisp, but used by RPRINT
(defun explode2 (x)
  (explode1 x nil))

;;;Not Standard Lisp, but used by PRETTY
(defun explodec (x)
  (explode1 x nil))

(defun explode1 (x escape-p)
  (cond ((symbolp x)
	 (let* ((name (symbol-name x))
		(clist '()))
	   (declare (list clist))
	   (when (and (digit-char-p (char name 0))
		      escape-p)
	     (push '! clist))
	   (dotimes (i (lisp:length name))
	     (let ((ch (char name i)))
	       (when (and escape-p
			  (not (or (digit-char-p ch)
				   (alpha-char-p ch))))
		 (push '! clist))
	       (push (svref single-character-symbols (char-code ch))
		     clist)))
	   (nreverse clist)))
	((stringp x)
	 (let ((clist '()))
	   (declare (list clist))
	   (when escape-p
	     (push '\" clist))
	   (dotimes (i (lisp:length x))
	     (let ((ch (char x i)))
	       (when (and (char= ch #\")
			  escape-p)
		 (push '\" clist))
	       (push (svref single-character-symbols (char-code ch))
		     clist)))
	   (when escape-p
	     (push '\" clist))
	   (nreverse clist)))
	((integerp x)
	 (if (eql x 0)
	     (list '\0)
	     (let ((clist '())
		   (minus nil))
	       (declare (list clist))
	       (when (minusp x)
		 (setq minus t)
		 (setq x (- x)))
	       (loop
		 (multiple-value-bind (q r)
		     (truncate x 10)
		   (push (svref single-character-symbols
				(+ (char-code #\0) r))
			 clist)
		   (when (eql q 0)
		     (return))
		   (setq x q)))
	       (when minus
		 (push '- clist))
	       clist)))
	(t (lisp:map 'list
		     #'(lambda (ch) (svref single-character-symbols
					   (char-code ch)))
		     (with-output-to-string (str)
		       (if escape-p
			   (prin1 x str)
			   (prin2 x str)))))))

;gensym (extended)

;;;Standard Lisp INTERN is required to take a symbol and intern it,
;;; not the string.  By popular demand, we punt on that requirement
;;; and just intern the symbol name.
(defun intern (x)
  (lisp:intern (if (symbolp x) (symbol-name x) x)))

;;; This is more correct, but causes problems for some implementations.
;;;
;;;  (defun intern (x)
;;;    (if (stringp x)
;;;        (lisp:intern x)
;;;        (progn (check-type x symbol)
;;;               (let ((name (symbol-name x)))
;;;                 (multiple-value-bind (name found) (find-symbol name)
;;;                   (if found
;;;                       name
;;;                       (progn (import x)
;;;                              x)))))))

(define-synonym remob unintern)


;3.4

;;;Reduce depends on GET or FLAG returning NIL when GETing on things
;;; other than symbols
(define-inline get (symbol indicator)
  (and (symbolp symbol) (lisp:get symbol indicator)))

(defun flag (list indicator)
  (when (symbolp indicator)
    (dolist (item list)
      (when (symbolp item)
	(setf (lisp:get item indicator) t)))))

(define-synonym flagp get (x y))

(defun put (symbol indicator value)
  (and (symbolp symbol)
       (symbolp indicator)
       (setf (lisp:get symbol indicator) value)))

(defun remflag (list indicator)
  (dolist (item list) (lisp:remprop item indicator)))

;;;Standard Lisp requires REMPROP to return what GET would return, while
;;; Common Lisp only requires some non-NIL value.
;;; (defun remprop (symbol indicator)
;;;   (and (symbolp symbol)
;;;        (do ((pl (symbol-plist symbol) (cddr pl))
;;;             (previous nil pl))
;;;            ((atom pl) nil)
;;;          (when (eq (car pl) indicator)
;;;            (if previous
;;;                (setf (cddr previous) (cddr pl))
;;;                (setf (symbol-plist symbol) (cddr pl)))
;;;            (return (cadr pl))))))

(defun remprop (sym indicator)
    (let ((ans (get sym indicator)))
	 (lisp:remprop sym indicator)
	 ans))

;3.5

(defvar *comp nil)

(defmacro de (name arglist &body body)
 #+HiLISP
  `(progn
     (eval-when (eval)
       (compdef ',name
		(lisp:function (lambda ,arglist ,@body))
		'(lambda ,arglist ,@body)))
     (eval-when (load compile)
       (defun ,name ,arglist ,@body))
     ',name)
 #-HiLISP
  `(progn
     (eval-when (eval)
       (compdef ',name
		(lisp:function (lambda ,arglist ,@body))
		'(lambda ,arglist ,@body)))
     (eval-when (load)
       (defun ,name ,arglist ,@body))
     ',name))

(defun compdef (name fun lambda-form)
  (if *comp
      (compile name lambda-form)
      (setf (symbol-function name) fun)))

;;;Common Lisp doesn't allow FEXPRs at all, but we simulate
;;;them with macros. EVALs inside them will not get the correct
;;;lexical environment.
(defmacro df (name (parameter) &body body)
  (let ((expr-name (lisp:intern (concatenate 'string
					(symbol-name name)
					(symbol-name '-expr)))))
    `(progn
       (defmacro ,name (&rest x)
	 `(,',expr-name ',x))
       (defun ,expr-name (,parameter)
	 ,@body)
       ',name)))

(defmacro dm (name (parameter) &rest body)
  `(defmacro ,name (&whole ,parameter &rest ignore)
     (declare (ignore ignore))
     ,@body))

(defun getd (x)
  (let ((mdef (macro-function x)))
    (cond (mdef
	   (if (member x '(prog return and or cond))
	       `(fexpr . ,mdef)
	       `(macro . ,#'(lambda (form)
			      (funcall mdef form nil)))))
	  ((special-form-p x)
	   `(fexpr . ,(symbol-function x)))	;This probably won't work.
	  ((fboundp x)
	   `(expr . ,(symbol-function x)))
	  (t nil))))

;;;Although DF "works", no provision is made for a direct PUTD of a FEXPR.

(defun putd (name type fn)
  (let ((interpreted-p nil)
	(fun fn))
    (when (consp fn)
      (setq fun (eval `(lisp:function ,fn)))
      (setq interpreted-p t))
    (cond ((eq type 'expr)
	   (if (and *comp interpreted-p)
	       (compile name fn)
	       (setf (symbol-function name) fun)))
	  ((eq type 'macro)
	   (when (and *comp interpreted-p)
	     (setq fun (compile nil fn)))
	   (setf (macro-function name)
		 #+OldSymbolics        ;;; Used to include VaxLisp
		 fun
		 #-(or OldSymbolics)
		 #'(lambda (form env)
		     (declare (ignore env))
		     (funcall fun form))))
	  (t (lisp:error "Illegal type ~S for PUTD of ~S" type name))))
  name)
	  
(defun remd (name)
  (if (fboundp name)
      (prog1 (getd name)
	     (fmakunbound name))
      nil))

;3.6

(defmacro fluid (var-list)
  (if (and (consp var-list) (eq (first var-list) 'quote))
      `(progn
	 (proclaim '(special ,@(second var-list)))
	 (setup-vars ,var-list 'fluid))
      `(setup-vars ,var-list 'fluid)))

(defun setup-vars (var-list type)
  (flag var-list type)
  (proclaim `(special ,@var-list))
  (dolist (var var-list)
    (if (not (boundp var))
	(setf (symbol-value var) nil))))

(defun fluidp (x)
  (and (symbolp x)
       (flagp x 'fluid)))

(defmacro global (var-list)
  (if (and (consp var-list) (eq (first var-list) 'quote))
      `(progn 'compile
	      (proclaim '(special ,@(second var-list)))
	      (setup-vars ,var-list 'global))
      `(setup-vars ,var-list 'global)))

(defun globalp (x)
  (and (symbolp x)
       (or (flagp x 'global) (fboundp x))))

;set

;setq (extended)

(defmacro unfluid (var-list)
  ;;There is no defined way in Common Lisp to
  ;; proclaim unspecial.
  (if (and (consp var-list) (eq (first var-list) 'quote))
      `(progn
	 #+(or LispM Lucid Excl)
	 (dolist (x ,var-list)
	   #-NewLucid
	   (remprop x 'special)
	   #+NewLucid
	   (lucid::proclaim-unspecial x))
	 #+VaxLisp
	 (dolist (x ,var-list) (remprop x 'system::globally-special))
	 (remflag ,var-list 'fluid))
      `(remflag ,var-list 'fluid)))



;3.7

;go (extended)

;prog (extended)

;progn

;prog2 (extended)

;return



;3.8

(defvar *message-p* t)

(defvar emsg*)

(defun error (number message)
  (setf emsg* message)
  (when *message-p*
    (cerror "Throw to ERRORSET"
	    "*****~:[ ~A~;~{ ~A~}~] (~A)"
	    (listp message) message number))
  (throw 'errorset number))

(defun errorset (form *message-p* trace-p)
  (declare (ignore trace-p))			;No backtrace is ever printed
  trace-p
  (catch 'errorset
    #+Symbolics
    (scl:catch-error-restart (zl:error "Return to Errorset ~S" form)
      (list (eval form)))
    #+HP
    (let ((return-value nil))
      (extn:when-error 
	  (setq return-value (list (eval form)))
        (format t "~a ~%" (extn:exception-msg))
	))
    #-(or HP Symbolics)
    (let ((return-value nil))
      (unwind-protect
	  (setq return-value (list (eval form)))
	(return-from errorset return-value)))
    ))



;3.9

(define-synonym getv svref (x y))

(define-inline mkvect (upper-bound)
  (make-array (the fixnum (1+ upper-bound)) :initial-element nil))

(define-inline putv (vector index element)
  (setf (svref vector index) element))

(define-inline upbv (vector)
  (the fixnum (1- (lisp:length (the simple-vector vector)))))


;3.10

;and

;cond (extended)

;not

;or



;3.11

;abs

(define-synonym add1 1+)

(define-synonym difference - (x y))

(defun divide (x y)
  (multiple-value-bind (q r)
      (truncate x y)
    (cons q r)))

;expt

(define-synonym fix floor)

;float

(define-synonym greaterp > (x y))

(define-synonym lessp < (x y))

;max

(define-synonym max2 max (x y))

;min

(define-synonym min2 min (x y))

(define-synonym minus -)

(defmacro plus (&rest numbers)
  `(+ ,@numbers))

(define-synonym plus2 + (x y))

(defun quotient (x y)
  (if (and (integerp x) (integerp y))
      (truncate x y)
      (/ x y)))

(define-synonym remainder rem (x y))

(define-synonym sub1 1-)

(defmacro times (&rest numbers)
  `(* ,@numbers))

(define-synonym times2 * (x y))


;3.12

(define-compiler-macro-and-function map (list function)
  `(progn (mapl ,function ,list)
	  nil))

(define-compiler-macro-and-function mapc (list function)
  `(progn (lisp:mapc ,function ,list)
	  nil))

(eval-when (compile eval)
(defmacro define-all-reversed-arguments (&rest function-names)
  `(progn
     ,@(lisp:mapcar
	 #'(lambda (name)
	     (let ((lisp-name
		     (lisp:intern (symbol-name name) (find-package "LISP"))))
	       `(define-reversed-arguments ,name ,lisp-name)))
	 function-names)))
)

(define-all-reversed-arguments mapcan mapcar mapcon maplist)



;3.13

;append (extended)

;;;Standard Lisp ASSOC uses EQUAL, not EQL.
(defun assoc (item a-list)
  (if (atom a-list)
      nil
      (let ((head (first a-list)))
	(if (or (atom head)
		(not (equal (car head) item)))
	    (assoc item (rest a-list))
	    head))))

(defun deflist (symbol-list indicator)
  (lisp:mapcar #'(lambda (x)
		   (let ((name (first x))
			 (value (second x)))
		     (put name indicator value)
		     name))
	       symbol-list))

(defun delete (item list)
  (if (atom list)
      list
      (let ((head (first list))
	    (tail (rest list)))
	(if (equal item head)
	    tail
	    (cons head (delete item tail))))))

(defun digit (x)
  (and (symbolp x)
       (let ((y (symbol-name x)))
	 (and (= (lisp:length y) 1)
	      (digit-char-p (char y 0))))))

;;;Standard Lisp LENGTH only applies to lists.
(defun length (l)
  (do ((i 0 (+ i 1))
       (ll l (rest ll)))
      ((atom ll) i)
    (declare (fixnum i))))

(defun liter (x)
  (and (symbolp x)
       (let ((y (symbol-name x)))
	 (and (= (lisp:length y) 1)
	      (alpha-char-p (char y 0))))))

;;;Changed to use EQL and keywords in Common Lisp
(defun member (item in-list)
  (cond ((atom in-list)
	 nil)
	((equal item (first in-list))
	 in-list)
	(t (member item (rest in-list)))))

#-(or LispM Lucid Spice Excl)
(defun memq (item in-list)
  ;;not Common Lisp, but still around in many implementations
  (cond ((atom in-list)
	 nil)
	((eq item (first in-list))
	 in-list)
	(t (memq item (rest in-list)))))

;nconc (extended)
#+(and Excl DG)
(defun nconc (u v)
  ;; If FOO is an atom, (NCONC NIL FOO) evaluates to FOO in Standard
  ;; Lisp, but to NIL in Franz Extended Common Lisp.
  (if (null u)
      v
      (lisp:nconc u v)))

(defun pair (x y)
  (lisp:mapcar #'cons x y))

;reverse (extended to sequences)

(defun sassoc (item list fn)
  (or (assoc item list)
      (funcall fn)))

(defun sublis (alist form)
  (if (atom alist)
      form
      (let ((found (assoc form alist)))
	(cond (found (cdr found))
	      ((atom form) form)
	      (t (let* ((head (car form))
			(tail (cdr form))
			(subhead (sublis alist head))
			(subtail (sublis alist tail)))
		   (if (and (eq head subhead)
			    (eq tail subtail))
		       form
		       (cons subhead subtail))))))))

;;;Standard Lisp SUBST uses EQUAL, not EQL
(defun subst (new old form)
  (cond ((equal old form) new)
	((atom form) form)
	(t (let* ((head (car form))
		  (tail (cdr form))
		  (subhead (subst new old head))
		  (subtail (subst new old tail)))
	     (if (and (eq head subhead) (eq tail subtail))
		 form
		 (cons subhead subtail))))))
 

;3.14

;apply (extended)

;eval

(defun evlis (x)
  (lisp:mapcar #'eval x))

(defun expand (l fn)
  (if (null (cdr l))
      (car l)
      `(,fn ,(car l) ,(expand (cdr l) fn))))

(defmacro function (x)
  (if (symbolp x)
      `',x
      `(lisp:function ,x)))

;quote



;3.15

;;;Regarding PAGELENTH/LINELENGTH/POSN/LPOSN:
;;; No attempt has been made to deal with the PAGELENGTH/LPOSN feature of
;;; Standard Lisp.  Hooks have been provided to force automatic line breaks
;;; in printing objects, but it was not felt necessary to use them.
;;; POSN returns
;;; 0 always except for streams which handle the :READ-CURSORPOS message.
;;; An association list has been used for LINELENGTH, which, although it does
;;; not affect printing here, is used by Reduce in printing.

(defun eject ()
  (write-char #\Page))				;Semi-standard character name

(defvar *stream-linelengths* nil "A-list (stream . linelength)")

(defconstant default-linelength 80.)

(defun linelength (x)
  ;;First, see if we've already got the stream in our table
  (let ((current-linelength
	  (lisp:assoc *standard-output* *stream-linelengths* :test #'eq)))
    ;;If not, then add it
    (when (null current-linelength)
      (setf current-linelength
	    (cons *standard-output*
		  ;;If the stream handles the :SIZE-IN-CHARACTERS message,
		  ;;use that.
		  (or #+LispM
		      (funcall *standard-output*
			       :send-if-handles :size-in-characters)
		      ;;Otherwise, use the default.
		      default-linelength)))
      ;;Finally, add the entry to the table
      (push current-linelength *stream-linelengths*))
    ;;OLD-LINELENGTH is the value returned by this function, no matter what.
    (let ((old-linelength (cdr current-linelength)))
      ;;If the argument was NIL, just return the value
      (when (not (null x))
	;;Otherwise, set the table entry to the new value (the argument)
	(setf (cdr current-linelength) x))
      ;;Finally, return the old value
      old-linelength)))
      

;;;No attempt has been made to include this feature.
(defun lposn ()
  0)

(defun open (file how)
  ;;HOW is either INPUT or OUTPUT
  (cond ((eq how 'INPUT)
	 (lisp:open file :direction :input))
	((eq how 'OUTPUT)
	 (if (probe-file file) 
	     (if (y-or-n-p "File ~S already exists. Supersede? " file)
		 (lisp:open file :direction :output
			    :if-exists :supersede))
	     (lisp:open file :direction :output)))
	(t (lisp:error "***** Incorrect argument to OPEN"))))
 

;The old open
;(defun open (file how)
;  ;;HOW is either INPUT or OUTPUT
;  (let ((stream (lisp:open file
;			   :direction
;			   (or (cdr (lisp:assoc how '((input . :input)
;						      (output . :output))))
;			       ;;Let LISP:OPEN give an error otherwise
;			       how))))
;    stream))

(defun close (stream)
  (setq *stream-linelengths*
	(lisp:delete stream *stream-linelengths* :key #'car))
  (lisp:close stream))

;;;No attempt has been made to include this feature.
(defun pagelength (x)
  (declare (ignore x))
  x	;Ignored
  0)

(defun posn ()
  (or #+LispM
      (funcall *standard-output* :send-if-handles :read-cursorpos :character)
      0))

(defun prin1 (x &optional (stream *standard-output*))
  (prin1-2 x t stream)
  x)

(defun prin2 (x &optional (stream *standard-output*))
  (prin1-2 x nil stream)
  x)

(defun princ (x &optional (stream *standard-output*))
  (prin1-2 x nil stream)
  x)

(defun print (x &optional (stream *standard-output*))
  (prin1 x stream)
  (terpri stream)
  x)

;;;These four functions provide a hook for the printer to cause automatic
;;; line breaks.  They are not used now because it's not worth the trouble
;;; and Reduce gets along fine without it.

(eval-when (eval compile)

(defmacro print-character (ch stream)
  `(write-char ,ch ,stream))

(defmacro print-string-contents (string stream)
  `(write-string ,string ,stream))

(defmacro whitespace (stream)
  `(write-char #\Space ,stream))

(defmacro token-done ()
  nil)

(defmacro print-anything (object slashify-p stream)
  (let ((obj (gensym))
	(str (gensym)))
    `(let ((,obj ,object)
	   (,str ,stream))
       (if ,slashify-p
	   (lisp:prin1 ,obj ,str)
	   (lisp:princ ,obj ,str)))))

)

(defun prin1-2 (object slashify-p stream)
  (cond ((consp object)
	 (prin1-2-cons object slashify-p stream))
	((symbolp object)
	 ;;This should do something about package prefixes,
	 ;; but I'm not sure what.
	 (let* ((name (symbol-name object))
		(len (lisp:length name)))
	   (if slashify-p
	       (progn (if (= len 0)
			  (print-string-contents "||" stream)
			  (if (digit-char-p (char name 0))
			      (print-character #\! stream)))
		      (dotimes (i len)
			(let ((ch (char name i)))
			  (when (not (or #-Excl
					 (upper-case-p ch)
					 #+Excl
					 (alpha-char-p ch)
					 (digit-char-p ch)))
			    (print-character #\! stream))
			  (print-character ch stream))))
	       (print-string-contents name stream))
	   (token-done)))
	((arrayp object)
	 (if (eql (array-rank object) 1)
	     (typecase object
	       (string
		 (if slashify-p
		     (progn (print-character #\" stream)
			    (dotimes (i (lisp:length object))
			      (let ((ch (char object i)))
				(when (char= ch #\")
				  (print-character #\" stream))
				(print-character ch stream)))
			    (print-character #\" stream))
		     (print-string-contents object stream)))
	       ((vector t)
		(prin1-2-vector object slashify-p stream))
	       (otherwise (print-anything object slashify-p stream)))
	     (print-anything object slashify-p stream)))
	(t (print-anything object slashify-p stream))))

(defun prin1-2-cons (object slashify-p stream)
  (print-character #\( stream)
  (token-done)
  (prin1-2 (car object) slashify-p stream)
  (do ((list (cdr object) (cdr list)))
      ((atom list)
       (when (not (null list))
	 (whitespace stream)
	 (print-character #\. stream)
	 (whitespace stream)
	 (prin1-2 list slashify-p stream)))
    (whitespace stream)
    (prin1-2 (first list) slashify-p stream))
  (print-character #\) stream)
  (token-done))

(defun prin1-2-vector (object slashify-p stream)
  (let ((len (lisp:length object)))
    (print-character #\[ stream)
    (when (> len 0)
      (token-done)
      (prin1-2 (svref object 0) slashify-p stream)
      (do ((i 1 (+ i 1)))
	  ((>= i len))
	(whitespace stream)
	(prin1-2 (svref object i) slashify-p stream)))
    (token-done)
    (print-character #\] stream)
    (token-done)))

(defun rds (stream)
  (when (null stream)
    (setf stream *terminal-io*))
  (prog1 *standard-input*
	 (setf *standard-input* stream)))

;read

(defvar $eof$ '$eof$)

#+Excl
(defvar *raise t)

#+Excl
(defvar *echo nil)

(defun readch ()
  (let ((ch (read-char *standard-input* nil $eof$)))
    (if (eq ch $eof$)
	ch
	#+(or Excl kcl)
	(progn
	 (if *echo (princ ch))
	 (svref single-character-symbols
		(char-code (if *raise (char-upcase ch) ch))))
	#-(or Excl kcl)
	(svref single-character-symbols (char-code ch)))))

;terpri

(defun wrs (stream)
  (when (null stream)
    (setf stream *terminal-io*))
  (prog1 *standard-output*
	 (setf *standard-output* stream)))


;3.16

;4

(defvar $eol$ '\
)

;;;*GC is not implemented.
(defvar *gc nil)

;nil

#-Excl
;;;*RAISE is not implemented.
(defvar *raise t)


;t

